#include <iostream>
#include <typeinfo>

class NullPointer
{
private:
   int line;
   const char *file;
   const char *type;

public:
   NullPointer(int a_line, const char *a_file, const char *a_type);
   void write(std::ostream &os) const;
};

NullPointer::NullPointer(int a_line, const char *a_file, const char *a_type)
   :
line(a_line), file(a_file), type(a_type)
{}

void NullPointer::write(std::ostream &os) const
{
   os<<"Null pointer exception encountered";
   //This is one of those situations where the pointer isn't REQUIRED
   if (file!=NULL)
   {
      os<<" at "<<file<<": "<<line<<",";
   }
   if (type !=NULL)
   {
      os<<" dereferencing pointer of type "<<type;
   }
   os<<"."<<std::endl;
}

// We could use this instead of the inline function, except that this way we
// can specialize for type
#define deref(x) deref2(x, __LINE__, __FILE__)

template <typename T>
inline T &deref2(T *ptr, int line=0, const char *file=NULL)
{
   if (ptr==NULL) throw NullPointer(line, file, typeid(ptr).name());
   return *ptr;
}

// Specialization for const char
inline const char &deref2(const char *ptr, int line=0, const char *file=NULL) 
{
   if (ptr==NULL) throw NullPointer(line, file, "const char *");
   return *ptr;
}

// Specialization for char
inline char &deref2(char *ptr, int line=0, const char *file=NULL)
{
   if (ptr==NULL) throw NullPointer(line, file, "char *");
   return *ptr;
}

// Specialization for NullPointer
inline NullPointer &deref2(NullPointer *ptr, int line=0, const char *file=NULL)
{
   if (ptr==NULL) throw NullPointer(line, file, "NullPointer *");
   return *ptr;
}

template <typename T>
std::ostream &operator << (std::ostream &os, T &writer)
{
   writer.write(os);
   return os;
}

int main()
{
   try
   {
      deref((const char *)NULL);
   }
   catch (NullPointer &err)
   {
      std::cout<<err;
   }
}
